package net.pricefx.common.apibuilder.quote;

import java.util.Map;

import net.pricefx.common.api.HeaderPhase;
import net.pricefx.common.apibuilder.IdGenerator;
import net.pricefx.common.apibuilder.Invocations;
import net.pricefx.common.apibuilder.MethodInvocation;
import net.pricefx.common.apibuilder.clicoll.CalculableLineItemCollectionBuilder;

/**
 * QuoteBuilder is the starting point of an interface that allows you manipulate
 * a quote object in a quote header calculation logic.<br />
 * The header calculation is executed twice: Once before the calculation of the
 * individual line items (= pre phase) and once after that (=post phase).<br />
 * Usually code should only apply in one phase. Use <code>isPrePhase()</code>
 * and <code>isPostPhase()</code> methods to detect current phase. <br />
 * The quote builder is available in the header logic by this call:<br />
 * <code>def builder = quoteProcessor</code>
 * <br /><br />
 * IMPORTANT: Please read super class notes on sequence of execution!
 *
 */
public class QuoteBuilder extends CalculableLineItemCollectionBuilder<QuoteBuilder> {

	// read only state, don't get propagated further, it is here for user convenience
	private final Object quoteView;
	private final QuoteHelper helper;
	
	@SuppressWarnings("unchecked")
	QuoteBuilder(String instanceId, String classId, IdGenerator idGenerator, Invocations<MethodInvocation> invocations, Object quoteView, HeaderPhase phase) {
		super(instanceId, classId, idGenerator, invocations, phase);
		this.quoteView = quoteView;
		this.helper = new QuoteHelper((Map<String, Object>) quoteView);
	}

	/**
	 * Gets the full quote object as a nested map. This map contains all header inputs/outputs, and all line items (with their inputs/outputs).
	 * 
	 * @return The quote object
	 */
	public Object getQuoteView() {		
		addInvocation("getQuoteView");		
		return quoteView;
	}
	
	/**
	 * Gets a helper class that assists in manipulating the quote object. You can use it instead of directly reading the QuoteView map.
	 * @return The helper object,
	 */
	public QuoteHelper getHelper() {
		addInvocation("getHelper");		
		return helper;
	}
	
	/**
	 * Apply a quote structure. That usually means adding a set of folders and/or items
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 * if (quoteProcessor.isPrePhase()) {
	 *      def structure = new QuoteStructure()
	 *      productSubCategory.each {
	 *           structure.addFolder(it.attribute2)
	 *      }
	 *      quoteProcessor.<strong>addQuoteStructure(structure)</strong>
	 * }
	 * </pre></blockquote>
	 *
	 * @param structure The structure
	 * @return
	 */
	public QuoteBuilder addQuoteStructure(QuoteStructure structure) {
		addInvocation("addQuoteStructure",structure);
		return this;
	}	
	
	/** 
	 * Creates and adds a new folder under the ROOT folder.
	 * <p>
	 * If you intend to add many folders and items, please use the QuoteStructure approach instead.
	 *
	 * @param label The folder label
	 */
	public void addFolder(String label) {

		QuoteStructure qs = new QuoteStructure();
		qs.addFolder(label);
		addQuoteStructure(qs);

	}	
	
	/**
	 * Creates and adds a new line item to the Quote with the given {@code sku}.
	 * <p>
	 * If you intend to add many folders and items, please use the QuoteStructure approach instead.
	 * @param sku The SKU of the new part
	 */
	public void addPart(String sku) {
		
		QuoteStructure qs = new QuoteStructure();
		qs.addPart(sku);
		addQuoteStructure(qs);
		
	}
	
	@Override
	public QuoteBuilderResult build() {
		addInvocation(METHOD_BUILD);

		QuoteBuilderResult r = new QuoteBuilderResult();
		r.invocations = invocations.get();
		return r;
	}

}