package net.pricefx.server.dto.calculation;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.googlecode.genericdao.search.Filter;

import net.pricefx.common.api.CalculationResultType;
import net.pricefx.common.api.FieldFormatType;
import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.logic.core.TypeCode;


/**
 * This class is used to transport a calculated matrix-style result
 * It is basically a list of maps with the addition that the column names are clearly named and can have a format
 * This means:
 * - the UI knows exactly which columns to display (even if some maps do not have those as keys)
 * - the entry maps can have more "columns" (or keys) than those that are displayed; maybe useful for calculation
 *
 * @author cat
 *
 */
@JsonIgnoreProperties(ignoreUnknown = true)
public class ResultMatrix implements ComplexCalculationResult, Serializable {

	private static final long serialVersionUID = 2266960694562305277L;

	private List<Map<String,Object>> entries;
	private List<String> columns;
	private FieldFormatType defaultFormat;
	private Map<String,FieldFormatType> columnFormats;
	private Map<String,String> columnTooltips;
	private boolean enableClientFilter;
	private String title;
	private String preferenceName;
    private boolean preferencesDisabled;
    private boolean sortingDisabled;
	private boolean escapeHTML;
	private Map<String,String> columnTranslations;

	private Map<String, Map<String, String>> onRowSelectEvents = new HashMap<>();
	
	@JsonProperty
	private List<Object> backEndActions = new ArrayList<>();

	public ResultMatrix() {
		this((Map<?, ?>) null);
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public ResultMatrix(Map jsonMap) {
		if(jsonMap == null) {
			entries = new ArrayList<Map<String,Object>>();
			columns = new ArrayList<String>();
		} else {
			entries = (List<Map<String, Object>>) jsonMap.get("entries");
			columns = (List<String>) jsonMap.get("columns");
		}

	}

	/**
	 * Gets the current title
	 * @return The current title
	 */
	public String getTitle() {
		return title;
	}

	/**
	 * Sets the title that will be shown as window title of the result matrix in the client
	 * @param title The title
	 */
	public void setTitle(String title) {
		this.title = title;
	}

	/**
	 * Sets the title that will be shown as window title of the result matrix in the client
	 * @param title The title
	 */
	public ResultMatrix withTitle(String title) {
		setTitle(title);
		return this;
	}

	/**
	 * Gets the preference name
	 * @return The preference name
	 */
	public String getPreferenceName() {
		return preferenceName;
	}

    /**
     * Tells whether saving of matrix grid preferences is disabled.
     *
     * @return true if saving of grid preferences is disabled. Default is false
     * @since 7.2 Bijou PR2
     */
    public boolean isPreferencesDisabled() {
        return preferencesDisabled;
    }

    /**
     * Allows to disable saving preferences of the matrix grid (columns, sorting, filtering). If not disabled, the
     * preferences may be be saved with the global dashboard preferences or attached to a specific preferences name
     * defined with setPreferenceName()/withPreferenceName()
     *
     * @param disablePreferences true means saving grid preferences is disabled. Default is false
     * @since 7.2 Bijou PR2
     */
    public void setDisablePreferences(boolean disablePreferences) {
        this.preferencesDisabled = disablePreferences;
    }

    /**
     * Allows to disable saving preferences of the matrix grid (columns, sorting, filtering). If not disabled, the
     * preferences may be be saved with the global dashboard preferences or attached to a specific preferences name
     * defined with setPreferenceName()/withPreferenceName()
     * 
     * @param disablePreferences true means saving grid preferences is disabled. Default is false
     * @since 7.2 Bijou PR2
     */
    public ResultMatrix withDisablePreferences(boolean disablePreferences) {
        setDisablePreferences(disablePreferences);
        return this;
    }

	/**
	 * UI sorting of matrix grid
	 *
	 * @return true if sorting is disabled on all columns. Default is false
	 */
		public boolean isSortingDisabled() {
		return sortingDisabled;
	}

	/**
	 * Set sorting capability in UI grid
	 * @param disableSorting true: sorting is disabled in UI on all columns
	 */
	public void setDisableSorting(boolean disableSorting) {
		this.sortingDisabled = disableSorting;
	}

	/**
	 * Set sorting capability in UI grid
	 * @param disableSorting true: sorting is disabled in UI on all columns
	 */
	public ResultMatrix withDisableSorting(boolean disableSorting) {
		setDisableSorting(disableSorting);
		return this;
	}

	/**
	 * Is HTML escaping enabled?
	 * @return true if the matrix should display raw text as is (i.e. not try to interpret it as HTML)
	 */
	public boolean isEscapeHTML() {
		return escapeHTML;
	}

	/**
	 * Set HTML escaping flag. Default: false (i.e. values will be interpreted as HTML)
	 * @param escapeHTML
	 */
	public void setEscapeHTML(boolean escapeHTML) {
		this.escapeHTML = escapeHTML;
	}

	/**
	 * Set HTML escaping flag. Default: false (i.e. values will be interpreted as HTML)
	 * @param escapeHTML
	 */
	public ResultMatrix withEscapeHTML(boolean escapeHTML) {
		setEscapeHTML(escapeHTML);
		return this;
	}

	/**
	 * Sets the so-called preference name. I.e. all result matrixes with the same preference name will share the same preference setting.
	 * The preference setting can be managed as usual in the client and can be set as global to predefine a good setting
	 *
	 * @param preferenceName The preference name
	 */
	public void setPreferenceName(String preferenceName) {
		if(preferenceName != null)
			this.preferenceName = preferenceName.replaceAll("[^a-zA-Z0-9]+","");
		else
			this.preferenceName = null;
	}

	/**
	 * Sets the so-called preference name. I.e. all result matrixes with the same preference name will share the same preference setting.
	 * The preference setting can be managed as usual in the client and can be set as global to predefine a good setting
	 *
	 * @param preferenceName The preference name
	 */
	public ResultMatrix withPreferenceName(String preferenceName) {
		setPreferenceName(preferenceName);
		return this;
	}

	/**
	 * Creates a new matrix object with an initial set of columns
	 * @param columns The initial set of column names
	 */
	public ResultMatrix(String ... columns) {
		this();
		for (int i = 0; i < columns.length; i++)
			addColumn(columns[i]);
	}

	/**
	 * Adds translated labels for the given column.
	 * @param column The column name (as defined e.g. in the constructor)
	 * @param translations A map with the translations. The key of the map is the locale, like "de".
	 */
	public void addColumnTranslation(String column, Map<String,String> translations) {
		if(columnTranslations == null) {
			columnTranslations = new HashMap<String,String>();
		}
		
//		columnTranslations.put(column, DomainObject.jsonSerialize(translations));
	}

	/**
	 * Adds translated labels for the given column.
	 * @param column The column name (as defined e.g. in the constructor)
	 * @param translations A map with the translations. The key of the map is the locale, like "de".
	 */
	public ResultMatrix withColumnTranslation(String column, Map<String,String> translations) {
		addColumnTranslation(column, translations);
		return this;

	}

	/**
	 * Gets current column translations in raw format
	 * @return
	 */
    public Map<String, String> getColumnTranslations() {
		return columnTranslations;
	}


	/**
     * Creates a new matrix object with an initial set of columns
     * @param columns Collection of column names
     */
	public ResultMatrix(Collection<String> columns) {
	    this(columns.toArray( new String[columns.size()]) );
	}

	/**
	 * Adds a column to the existing set of columns
	 * @param column The columns name
	 */
	public void addColumn(String column) {
		if(!columns.contains(column))
			columns.add(column);
	}

	/**
	 * Adds a column to the existing set of columns
	 * @param column The columns name
	 */
	public ResultMatrix withColumn(String column) {
		addColumn(column);
		return this;
	}

	/**
	 * Adds a column to the existing set of columns including a tooltip
	 * @param column The columns name
	 */
	public void addColumn(String column, String tooltip) {
		if(!columns.contains(column))
			columns.add(column);

		setColumnTooltip(column,tooltip);
	}

	/**
	 * Adds a column to the existing set of columns including a tooltip
	 * @param column The columns name
	 */
	public ResultMatrix withColumn(String column, String tooltip) {
		addColumn(column);
		setColumnTooltip(column,tooltip);
		return this;
	}

	/**
     * Adds columns to the existing set of columns
     * @param columns Collection of columns names
     */
	public void addColumns(Collection<String> columns) {
	    for (String col : columns){
	        addColumn(col);
	    }
	}

	/**
	 * Adds columns to the existing set of columns
	 * @param columns Collection of columns names
	 */
	public ResultMatrix withColumns(Collection<String> columns) {
		addColumns(columns);
		return this;
	}

	/**
	 * Adds a data row. The keys of the row-map should match the column names defined earlier
	 * @param row
	 */
	public void addRow(Map<String,Object> row) {
		entries.add(row);
	}

    /**
     * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
     * @param row Row as List
     */
   public void addRow(List<Object> row) {
       if (row != null){
           Map<String,Object> m = new HashMap<>();
           for (int c=0; c < Math.min(row.size(), columns.size()); c++){
               m.put(columns.get(c), row.get(c));
           }
           entries.add(m);
       }
    }

	/**
	 * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
	 * @param row Row as List
	 */
	public ResultMatrix withRow(List<Object> row) {
		addRow(row);
		return this;
	}

	/**
	 * Adds multiple data rows. The keys of the row-map should match the column names defined earlier
	 * @param rows Row as List of Maps
	 */
	public ResultMatrix withRows(List<Map<String,Object>> rows) {
		if(rows != null) {
			for(Map<String,Object> row : rows) {
				addRow(row);
			}
		}
		return this;
	}
   /**
    * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
    * @param values Row as Array
    */
   @SuppressWarnings("unchecked")
   public void addRow(Object... values) {
       if (values != null){
           if (values.length == 1 && values[0] instanceof Map) {
               addRow((Map<String,Object>) values[0]);
           }
           else{
               addRow(Arrays.asList(values));
           }
       }
   }

	/**
	 * Gets a list of defined columns
	 * @return The columns
	 */
	public List<String> getColumns() {
		return columns;
	}

	/**
	 * Returns the current list entries as a list (=the rows) of maps (= the row content)
	 * @return List of maps
	 */
	public List<Map<String,Object>> getEntries() {
		return entries;
	}

	/**
	 * Retrieves a particular row based on a value check
	 * @param key The column name to search on
	 * @param value The value to match the row's column value to
	 * @return The row
	 */
	public Map<String,Object> findRow(String key, Object value) {
		for(Map<String,Object> entry : entries) {
			Object val = entry.get(key);
			if(val == null && value == null) return entry;
			if(val != null && val.equals(value)) return entry;
		}
		return null;
	}

	/**
	 * Transpose the matrix.
	 *
	 * <br/><br/>
	 *
	 * Example:
	 * <PRE>
	 * ProductGroup | AvgPrice | AvgMargin &lt;- header row
	 * -----------------------------------
	 *     Beef     |   4.1€   |   3.6%
	 *     Pork     |   2.2€   |   1.1%
	 *     Lamb     |   4.6€   |   5.2%
	 * </PRE>
	 *
	 * Transposed matrix:
	 * <PRE>
	 * transpose("ProductGroup")
	 *
	 * ProductGroup | Beef | Pork | Lamb &lt;- header row
	 * ---------------------------------
	 *   AvgPrice   | 4.1€ | 2.2€ | 4.6€
	 *   AvgMargin  | 3.6% | 1.1% | 5.2%
	 * </PRE>
	 *
	 *
	 * @param newHeaderColumnName Name of the column that will become the new header row
	 * @return The transposed ResultMatrix (new instance)
	 */
	public ResultMatrix transpose(String newHeaderColumnName) throws Exception {
		throw new Exception("Method ResultMatrix.transpose() is not supported by debugger");
	}
	

	/**
	 * Transpose the matrix.
	 *
	 * <br/><br/>
	 *
	 * Example:
	 * <PRE>
	 * ProductGroup | AvgPrice | AvgMargin  &lt;- header row
	 * ------------------------------------
	 *     Beef     |   4.1€   |   3.6%
	 *     Pork     |   2.2€   |   1.1%
	 *     Lamb     |   4.6€   |   5.2%
	 * </PRE>
	 *
	 * Transposed matrix:
	 * <PRE>
	 * transpose("ProductGroup", "PG Label")
	 *
	 *  Beef | Pork | Lamb |   PG Label    &lt;- header row
	 * -----------------------------------
	 *  Beef | Pork | Lamb | ProductGroup
	 *  4.1€ | 2.2€ | 4.6€ |   AvgPrice
	 *  3.6% | 1.1% | 5.2% |   AvgMargin
	 * </PRE>
	 *
	 *
	 * @param newHeaderColumnName Name of the column that will become the new header row
	 * @param oldColLabelsInNewColumnName Name of the new column that will receive the old column headers as values (optional)
	 * @return The transposed ResultMatrix (new instance)
	 */
	public ResultMatrix transpose(String newHeaderColumnName, String oldColLabelsInNewColumnName) throws Exception {
		throw new Exception("Method ResultMatrix.transpose() is not supported by debugger");
	}

	public Matrix2D toMatrix2D() {
	    return Matrix2D.fromResultMatrix(this);
	}

	/**
	 * Sets a default column format
	 * @param fft
	 */
	public void setDefaultFormat(FieldFormatType fft) {
		defaultFormat = fft;
	}

	/**
	 * Sets a default column format
	 * @param fft
	 */
	public ResultMatrix withDefaultFormat(FieldFormatType fft) {
		setDefaultFormat(fft);
		return this;
	}

	/**
	 * Gets the default format
	 * @return Default format
	 */
	public FieldFormatType getDefaultFormat() {
		return defaultFormat;
	}

	/**
	 * Gets individual column formats (if specified)
	 * @return A Map with column name and format
	 */
	public Map<String, FieldFormatType> getColumnFormats() {
		return columnFormats;
	}

	public Map<String, Map<String, String>> getOnRowSelectEvents() {
		return onRowSelectEvents;
	}

	/**
	 * Sets a column format for a specified column
	 * @param column The column name
	 * @param fft The format
	 */
	public void setColumnFormat(String column, FieldFormatType fft) {
		if(columnFormats == null)
			columnFormats = new HashMap<String,FieldFormatType> ();
		columnFormats.put(column, fft);
	}

	/**
	 * Sets a column format for a specified column
	 * @param column The column name
	 * @param fft The format
	 */
	public ResultMatrix withColumnFormat(String column, FieldFormatType fft) {
		setColumnFormat(column,fft);
		return this;
	}

	/**
	 * Sets a column formats for multiple columns. Adds column if not existing yet
	 * @param columnFormats A key value map of column:format type
	 */
	public ResultMatrix withColumnFormats(Map<String,FieldFormatType> columnFormats) {
		if(columnFormats != null) {
			for (String key : columnFormats.keySet()) {
				addColumn(key);
				setColumnFormat(key,columnFormats.get(key));
			}
		}
		return this;
	}

	/**
	 * Gets the currently assigned column tooltips
	 * @return
	 */
	public Map<String, String> getColumnTooltips() {
		return columnTooltips;
	}

	/**
	 * Adds a tooltip to a column
	 * @param column The column name
	 * @param tooltip The tooltip text
	 */
	public void setColumnTooltip(String column, String tooltip) {

		if(columnTooltips == null)
			columnTooltips = new HashMap<String,String>();

		if(tooltip == null)
			columnTooltips.remove(column);
		else
			columnTooltips.put(column, tooltip);
	}

	/**
	 * Adds a tooltip to a column
	 * @param column The column name
	 * @param tooltip The tooltip text
	 */
	public ResultMatrix withColumnTooltip(String column, String tooltip) {
		setColumnTooltip(column,tooltip);
		return this;
	}

	public boolean isEnableClientFilter() {
		return enableClientFilter;
	}

	public void setEnableClientFilter(boolean enableClientFilter) {
		this.enableClientFilter = enableClientFilter;
	}

	public ResultMatrix withEnableClientFilter(boolean enableClientFilter) {
		setEnableClientFilter(enableClientFilter);
		return this;
	}

	/**
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @return The styled cell object that can then be further styled with e.g. the flow pattern .withXXX() methods
	 */
	public ResultMatrixStyledCell styledCell(Object value) {
		ResultMatrixStyledCell rmcc = new ResultMatrixStyledCell();
		rmcc.value = value;
		return rmcc;
	}

	/**
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @param textColor The text color of the cell
	 * @param bgColor The background color of the cell (HTML hex value)
	 * @return The styled cell object
	 */
	public ResultMatrixStyledCell styledCell(Object value, String textColor,String bgColor) {
		ResultMatrixStyledCell rmcc = new ResultMatrixStyledCell();
		rmcc.value = value;
		rmcc.color = textColor;
		rmcc.background = bgColor;
		return rmcc;
	}

	/**
	 *
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @param textColor The text color of the cell
	 * @param bgColor The background color of the cell (HTML hex value)
	 * @param weight The font weight
	 * @return The styled cell object
	 */
	public ResultMatrixStyledCell styledCell(Object value, String textColor,String bgColor, String weight) {
		ResultMatrixStyledCell rmsc = new ResultMatrixStyledCell();
		rmsc.value = value;
		rmsc.color = textColor;
		rmsc.background = bgColor;
		rmsc.fontweight = weight;
		return rmsc;
	}

	/**
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @param textColor The text color of the cell
	 * @param bgColor The background color of the cell (HTML hex value)
	 * @param weight The font weight
	 * @param alignment Cell alignement value (valid CSS)
	 * @return The styled cell object
	 */
	public ResultMatrixStyledCell styledCell(Object value, String textColor,String bgColor, String weight, String alignment) {
		ResultMatrixStyledCell rmsc = new ResultMatrixStyledCell();
		rmsc.value = value;
		rmsc.color = textColor;
		rmsc.background = bgColor;
		rmsc.fontweight = weight;
		rmsc.alignment = alignment;
		return rmsc;
	}

	/**
	 * Creates an image cell that can then be added as row value. Value should be a valid image URL
	 * @param value The image URL
	 * @return The image cell object
	 */
	public ResultMatrixImageCell imageCell(Object value) {
		ResultMatrixImageCell rmic = new ResultMatrixImageCell();
		rmic.value = value;
		return rmic;
	}

	public ResultMatrixImageCell libraryImage(String mainName, String statusName) {
		ResultMatrixImageCell rmic = new ResultMatrixImageCell();
		rmic.value = "images/library/" + mainName + "_" + statusName + ".png";
		return rmic;
	}

	public ResultMatrixLinkCell linkToPriceGrid(Object value, long priceGridID, Filter filter) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();

		rmlc.value = value;
		rmlc.targetPage = "priceGridPage";

        HashMap<String, Object> state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
                priceGridID + "." + TypeCode.PRICEGRID.getShortName());
        state.put(PriceFxInterface.PARAM_FILTER_CRITERIA, filter);

		rmlc.targetPageState = state;

		return rmlc;
	}

	public Object linkToPriceList(Object value, long priceListID, Filter filter) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();

        rmlc.value = value;
        rmlc.targetPage = "priceListPage";

		HashMap<String, Object> state = new HashMap<String, Object>(2);
		state.put(PriceFxInterface.FIELD_ID,
				priceListID + "." + TypeCode.PRICELIST.getShortName());
		state.put(PriceFxInterface.PARAM_FILTER_CRITERIA, filter);

		rmlc.targetPageState = state;

		return rmlc;
	}

	public ResultMatrixLinkCell linkToPriceParameter(Object value, long priceParameterID) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();

		rmlc.value = value;
		rmlc.targetPage = "pricingParametersPage";

        HashMap<String, Object> state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
				priceParameterID + "." + TypeCode.LOOKUPTABLE.getShortName());

		rmlc.targetPageState = state;

		return rmlc;
	}

	public ResultMatrixLinkCell linkCell(Object value, String targetPage, String targetPageState) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();
		rmlc.value = value;
		rmlc.targetPage = targetPage;
		rmlc.targetPageState = targetPageState;
		return rmlc;
	}

	public ResultMatrixBackEndCell backEndAction(String label, String url, String payload, String successMessage, String failMessage) {
		ResultMatrixBackEndCell cell = new ResultMatrixBackEndCell();
		cell.label = label;
		cell.values = new String[] {url, payload, successMessage, failMessage};
		return cell;
	}

    /**
     * <p>Allows UI users to select multiple rows and adds an action button next to the ResultMatrix. Multiple such
     * action buttons can be defined, identified by the provided <code>id String</code>.</p>
     * <p>The action will be handled by a dedicated logic, which receives an <code>InputMatrix</code> parameter
     * named after the provided <code>id</code> and featuring the row values of a subset of the columns.</p>
     * <p>Example:</p>
     * <pre>
     *     def matrix = api.newMatrix("col1", "col2", "col3")
     *     matrix.addRow(1, "a", "f")
     *     matrix.addRow(2, "b", "g")
     *     matrix.rowSelectionBackEndAction("actionId")
     *                 .withLogicName("action_handler")
     *                 .withColumns("col2", "col3")
     *                 .withButtonLabel("Click here")
     *                 .withSuccessMessage("Success")
     *                 .withFailureMessage("Sorry")
     *     
     * </pre>
     * 
     * <p>Note: This is available in Unity UI only. No backporting planned for the Classic UI.</p>
     * 
     * See {@link ResultMatrixBackEndAction} to explore the full list of methods allowing to configure an action.
     * 
     * @param id  a <code>String</code> identifying the action
     * @return  a customizable object representing an action button, which the UI will then render and make functional
     * @since 7.0 Bijou
     */
    public Object rowSelectionBackEndAction(final String id) {
        final Object action = id;
        backEndActions.add(action);
        return action;
    }

    public ResultMatrixDownloadCell downloadButton(String label, String url, String payload) {
		ResultMatrixDownloadCell cell = new ResultMatrixDownloadCell();
		cell.label = label;
		cell.values = new String[] {url, payload};
		return cell;
	}

	public ResultMatrixCells cells(String label, Object... cells) {
		ResultMatrixCells cs = new ResultMatrixCells();
		cs.label = label;
		cs.cells = cells;
		return cs;
	}

	/**
	 * Configures a client-side UI event. Use case is usually a refresh of some sort when a result matrix row is selected
	 * Example code:
	 * <pre>
	 * matrix.onRowSelection().triggerEvent(api.dashboardWideEvent("CustomerIdChanged-ED")).withColValueAsEventDataAttr("CustomerID", "customerId")
	 * </pre>
	 * The embedded dashboard that receives the event could then look like:
	 * <pre>
	 *  return api.dashboard("EmbeddedDashboard")
	 *
	 *    // Here the user can assign values to the embedded
	 *    // dashboard inputs common to all three modes
	 *    .setParam("Customer Id", api.input("Customer Id"))
	 *    .setParam("Year", api.input("Year"))
	 *
	 *    // Show the dashboard embedded ...
	 *    .showEmbedded()
	 *    // ... and reevaluate it on "CustomerIdChanged-ED"
	 *    // Note: "dashboardWideEvent()" fce makes the event local
	 *    // to the containing dashboard instance
	 *    .andRecalculateOn(api.dashboardWideEvent("CustomerIdChanged-ED"))
	 *    // Pull the "customerId" attribute out of the event payload and expose it
	 *    // as the "Customer Id" input to the embedded dashboard
	 *    .withEventDataAttr("customerId").asParam("Customer Id")
	 * </pre>
	 * @return The selection event to further configure it in a flow pattern
	 */
	public ResultMatrixOnRowSelection onRowSelection() {
		return new ResultMatrixOnRowSelection();
	}

	@Override
	public String toString() {
		return String.format("(%s),(%s)", columns,entries);
	}

	@Override
	public String getResultType() {
		return CalculationResultType.MATRIX.name();
	}

	protected FieldFormatType decideFormat(Object input) {
		if (input instanceof org.joda.time.LocalDate ||
				input instanceof java.time.LocalDate)
		{
			return FieldFormatType.DATE;
		}
		if (input instanceof org.joda.time.DateTime ||
				input instanceof java.time.LocalDateTime ||
				input instanceof org.joda.time.LocalDateTime ||
				input instanceof java.util.Calendar ||
				input instanceof java.util.Date ||
				input instanceof java.time.Instant ||
				input instanceof org.joda.time.Instant) {
			return FieldFormatType.DATETIME;
		}

		if (input instanceof Integer) {
			return FieldFormatType.INTEGER;
		}
		if (input instanceof Long) {
			return FieldFormatType.NUMERIC_LONG;
		}

		if (input instanceof Number) {
			return FieldFormatType.NUMERIC;
		}

		if (input instanceof String) {
			return FieldFormatType.TEXT;
		}

		return null;
	}

	public void addFormatsToResultMatrix() {
		getEntries().stream() // one row is map of columns
				.flatMap(e -> e.entrySet().stream()) // stream of colums of all rows
				.filter(entry -> entry.getValue() != null);
//				.map(entry -> new Pair<>(entry.getKey(), decideFormat(entry.getValue())))
//				.filter( p -> p.getValue() != null)
//				.forEach( p-> setColumnFormat(p.getKey(), p.getValue()));
	}

	/**
	 * Search for non null <a href="#{@link}">{@link FieldFormatType#DATETIME}</a> columns and convert them into
	 * <a href="#{@link}">{@link org.joda.time.DateTime}</a>
	 */
	public void normalizeDateTimes() {
		Set<String> datetimeKeys = 	columnFormats.entrySet().stream()
				     	.filter( e -> e.getValue() == FieldFormatType.DATETIME)
				        .map(Map.Entry::getKey)
					 	.collect(Collectors.toSet());
		for (Map<String, Object> row : getEntries()) {
			for (Map.Entry<String, Object> column : row.entrySet()) {
				if (datetimeKeys.contains(column.getKey()) && column.getValue() != null) {
//					column.setValue(DateTimeUtil.convert(column.getValue()));
				}
			}
		}

	}


	public static abstract class AbstractResultMatrixCell {
		public Object value;
		public String type;
	}

	public static class ResultMatrixStyledCell extends AbstractResultMatrixCell {
		public String color;
		public String background;
		public String fontweight;
		public String alignment;
		public String extraCSS;
		public String tooltip;

		public ResultMatrixStyledCell() {
			type = "StyledCell";
		}

		/**
		 * Set the foreground color of the value font
		 * @param color
		 * @return
		 */
		public ResultMatrixStyledCell withColor(String color) {
			this.color = color;
			return this;
		}

		/**
		 * Set the background color of the cell
		 * @param bgColor
		 * @return
		 */
		public ResultMatrixStyledCell withBackground(String bgColor) {
			this.background = bgColor;
			return this;
		}

		/**
		 * Set the font weight (in standard CSS terms) of the value text
		 * @param fontWeight
		 * @return
		 */
		public ResultMatrixStyledCell withFontWeight(String fontWeight) {
			this.fontweight = fontWeight;
			return this;
		}

		/**
		 * Specify the alignment of the cell
		 * @param alignment
		 * @return
		 */
		public ResultMatrixStyledCell withAlignment(String alignment) {
			this.alignment = alignment;
			return this;
		}

		/**
		 * Show a tooltip text in the cell
		 * @param tooltip
		 * @return
		 */
		public ResultMatrixStyledCell withToolTip(String tooltip) {
			this.tooltip = tooltip;
			return this;
		}

		/**
		 * Apply any other valid CSS instructions to the cell
		 * @param extraCSS
		 * @return
		 */
		public ResultMatrixStyledCell withCSS(String extraCSS) {
			this.extraCSS = extraCSS;
			return this;
		}
	}

	public static class ResultMatrixLinkCell extends AbstractResultMatrixCell {
		public String targetPage;
		public Object targetPageState;

		public ResultMatrixLinkCell() {
			type = "LinkCell";
		}
	}

	/**
	 * Create a link/button that should do some desired action to pricefx backend (unity only)
	 * [url:"pricelistmanager.revoke..../...", payload: [... : ...], successMessage: "hey, it is ok!", failMessage: "ayay, this failed"]
	 */
	public static class ResultMatrixBackEndCell {
		public String type;
		public String label;
		public String[] values;
		public ResultMatrixBackEndCell() {
			type = "BackEndAction";
		}
	}

	/**
	 * Create a link/button that should do initiate download from pricefx backend (unity only)
	 * [url:"pricelistmanager.fetchpdf..../...", payload: [... : ...]]
	 */
	public static class ResultMatrixDownloadCell {
		public String type;
		public String label;
		public String[] values;
		public ResultMatrixDownloadCell() {
			type = "DownloadButton";
		}
	}

	/**
	 * List of other cells (to be shown as dropdown inline
	 */
	public static class ResultMatrixCells {
		public Object[] cells;
		public String label;
		public String type;
		public ResultMatrixCells() {
			type = "Cells";
		}
	}

	public static class ResultMatrixImageCell extends AbstractResultMatrixCell {

		public ResultMatrixImageCell() {
			type = "ImageCell";
		}
	}

	public class ResultMatrixOnRowSelection {

		public ResultMatrixOnRowSelection() {
		}

		public TriggerEvent triggerEvent(String eventName) {
			if (eventName == null || eventName.isEmpty()) {
				throw new IllegalArgumentException("Illegal argument in withColValueAsEventDataAttr(...)");
			}
			HashMap<String, String> mappings = new HashMap<String, String>();
			onRowSelectEvents.put(eventName, mappings);
			return new TriggerEvent(mappings);
		}

		public class TriggerEvent {

			private final Map<String, String> mappings;

			public TriggerEvent(HashMap<String, String> mappings) {
				this.mappings = mappings;
			}

            public TriggerEvent withColValueAsEventDataAttr(String colName, String attrName) {
                if (colName == null || colName.isEmpty()) {
                    throw new IllegalArgumentException("Illegal argument in withColValueAsEventDataAttr(...)");
                }
                if (attrName == null || attrName.isEmpty()) {
                    throw new IllegalArgumentException("Illegal argument in withColValueAsEventDataAttr(...)");
                }
                mappings.put(colName, attrName);
                return TriggerEvent.this;
            }
		}
	}

    /**
     * <p>Object representing a button that handles multi-row actions.</p>
     * <p>The action will be handled by a dedicated logic, which receives an <code>InputMatrix</code> parameter
     * named after the provided <code>id String</code> and featuring the row values of a subset of the columns.</p>
     * 
     * <p>Note: This is available in Unity UI only. No backporting planned for the Classic UI.</p>
     * 
     * See {@link ResultMatrix#rowSelectionBackEndAction(String)} to find out how to create an instance of
     * <code>ResultMatrixBackEndAction</code>.
     * 
     * @since 7.0 Bijou
     */
    public static class ResultMatrixBackEndAction implements Serializable {

        // we may want to add more types, but so far most useful one that handles it all is formulamanager.executeformula
        @JsonProperty
        private final String type = "formula";

        @JsonProperty
        private final String id;

        @JsonProperty
        private final LogicParameters parameters = new LogicParameters();

        private static class LogicParameters implements Serializable {
            @JsonProperty
            private String logicName;
            @JsonProperty
            private Date targetDate;
            @JsonProperty
            private final List<String> columns = new ArrayList<>();
            @JsonProperty
            private final Map<String, Object> additionalInputs = new HashMap<>();
        }

        // Visible elements in the UI / user feedback
        @JsonProperty
        private String buttonLabel;
        @JsonProperty
        private String buttonTooltip;
        @JsonProperty
        private String successMessage;
        @JsonProperty
        private String failureMessage;

        /**
         * For deserialization only, it's good news if your IDE tells you it's unused.
         */
        private ResultMatrixBackEndAction() {
            this(null);
        }

        private ResultMatrixBackEndAction(final String actionId) {
            this.id = actionId;
        }

        /**
         * @param name the logic to execute
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withLogicName(final String name) {
            this.parameters.logicName = name;
            return this;
        }

        /**
         * @param date <b>optional</b> targetDate in case there are multiple active versions with the same name
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withTargetDate(final Date date) {
            this.parameters.targetDate = date;
            return this;
        }

        /**
         * @param names list of column names that will be passed as <code>InputMatrix</code> to the logic
         *              (replaces any previously set column names)
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withColumns(final String... names) {
            this.parameters.columns.clear();
            this.parameters.columns.addAll(Arrays.asList(names));
            return this;
        }

        /**
         * @param name name to add in the list of columns that will be passed as <code>InputMatrix</code> to the logic
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction addColumn(final String name) {
            this.parameters.columns.add(name);
            return this;
        }

        /**
         * @param name  name of the input parameter that will be passed to the logic
         * @param value value of the input parameter that will be passed to the logic (keep in mind that this value makes
         *              a round trip to the front end in JSON format)
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction addFormulaInput(final String name, final Object value) {
            this.parameters.additionalInputs.put(name, value);
            return this;
        }

        /**
         * @param label the label of the action button
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withButtonLabel(final String label) {
            this.buttonLabel = label;
            return this;
        }

        /**
         * @param tooltip <b>optional</b> hover message for additional info about the action
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withButtonTooltip(final String tooltip) {
            this.buttonTooltip = tooltip;
            return this;
        }

        /**
         * @param text message to display in case of success of the backend call
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withSuccessMessage(final String text) {
            this.successMessage = text;
            return this;
        }

        /**
         * @param text message to display in case of failure of the backend call
         * @since 7.0 Bijou
         */
        public ResultMatrixBackEndAction withFailureMessage(final String text) {
            this.failureMessage = text;
            return this;
        }

    }

}
